const DB = require("../database/DB");
const Gzip = require("../functions/Gzip");
const SystemLogger = require("../logs/SystemLogger");
const AssistantConfig = require("../model/AssistantConfig");
const AssistantControl = require("../service/AssistantControl");
const moment = require('moment');

class AssistantConfigController {
  async checkForUpdates() {
    const connection = await DB.beginTransaction();
    try {
      const monitors = await AssistantControl.getAssistantMonitoring();
      const currentMonitors = await AssistantConfig.getAssistantConfig();
      const mappedLatestMonitors = {};

      if (!monitors) return currentMonitors;

      monitors.forEach(monitor => {
        const { name, monitoringId, content, description, interval, isActive} = monitor;
        mappedLatestMonitors[monitoringId] = {
          NOME: name,
          DESCRICAO: description,
          COMANDO: content || 'INTERNO',
          IDMONITORAMENTO: monitoringId,
          INTERVALO: interval,
          ATIVO: isActive,
          updateAt: new Date(monitor.updatedAt).getTime(),
          COMPLEMENTOTIPO: 'N',
          TIPO: 'I',
          ORDEM: 1
        };
      });
      const curMonitorKeys = [];
      currentMonitors.rows.forEach(curMonitor => {
        const { TIPO, IDMONITORAMENTO, DTULTATU } = curMonitor;
        if(TIPO === 'I') {
          curMonitorKeys.push(IDMONITORAMENTO);
          
          if(IDMONITORAMENTO in mappedLatestMonitors) {
            const updateAtCurMonitor = moment(DTULTATU, 'DD/MM/YYYY HH24:MI:SS')._d.getTime();
            const { updateAt } = mappedLatestMonitors[IDMONITORAMENTO] || {};
            
            if(updateAt > updateAtCurMonitor) {
              mappedLatestMonitors[IDMONITORAMENTO] = Object.assign(curMonitor, mappedLatestMonitors[IDMONITORAMENTO])
              mappedLatestMonitors[IDMONITORAMENTO].toUpdate = true;
            }
          } else {
            mappedLatestMonitors[IDMONITORAMENTO] = {
              toDelete: true,
              ID: curMonitor.ID,
            }
          }
        }
      });

      await Promise.all(Object.keys(mappedLatestMonitors).map(monitoringId => {
        const { toUpdate = false, toDelete = false, isNew = false, ...monitor} = mappedLatestMonitors[monitoringId];
        if(toUpdate) return AssistantConfig.update(monitor, { connection });
        if(toDelete) return AssistantConfig.delete(monitor, { connection });
        if(!curMonitorKeys.includes(monitor.IDMONITORAMENTO)) {
          return AssistantConfig.insert(monitor, { connection });
        }
        return Promise.resolve();
      }));
      await connection.commit();
    } catch(err) {
      SystemLogger.error(err);
      await connection.rollback();
    } finally {
      await connection.close();
    }
  }

  async list() {
    await this.checkForUpdates().catch(e => SystemLogger.error(e));
    const assistantConfig = await AssistantConfig.getAssistantConfig();
    if (!assistantConfig) return [];

    assistantConfig.rows.forEach(row => {
      row.COMANDO = Gzip.decode(row.COMANDO);
    });

    return assistantConfig.rows;
  }
}

module.exports = new AssistantConfigController();