const oracledb = require('oracledb');
require('moment/locale/pt-br');
const StoredProcedure = require('./StoredProcedure');
const Connection = require('./Connection');
const Parameter = require('./Parameter');
const DataBaseLogger = require('../../logs/DataBaseLogger');

class Oracle {
  constructor() {
    process.env.ORA_SDTZ = 'UTC';
    oracledb.fetchAsString = [ oracledb.CLOB ];
    oracledb.autoCommit = true;
  }

  async beginTransaction() {
    const conn = await Connection.connect({ useNewConnection: true });
    return conn;
  }

  async fetch(sql, params, fetchParams = { outputObj: true }) {
    try {
      const conn = await Connection.getActiveConnection();
      const query = Parameter.bindQueryParams(sql, params);
      const executionParams = {};
      if(fetchParams.outputObj) {
        executionParams.outFormat = oracledb.OUT_FORMAT_OBJECT;
      }
      return await conn.execute(query, [], executionParams);
    } catch (error) {
      DataBaseLogger.write(error, {type: 'QUERY_EXECUTION', sql, params})
      return Promise.reject(new Error('DB_ERROR'));
    }
  }

  async executeProcedure(name, procedureParams) {
    try {
      const conn = await Connection.getActiveConnection();
      const { sql, params } = StoredProcedure.prepare(name, procedureParams);
      return await conn.execute(sql, params);
    } catch (error) {
      return Promise.reject(new Error('DB_ERROR'));
    }
  }

  async executeAnonymousBlock(sql, columns) {
    try {
      const result = await conn.execute(`BEGIN ${sql}; END;`, columns);
      result.connection = conn;
      return result;
    } catch(error) {
      DataBaseLogger.write(error, {type: 'ANONYMOUS_BLOCK', sql, params: {}});
      return Promise.reject(new Error('DB_ERROR'));
    }
  }

  async execute(sqlString, params, { connection = null, outParams = [] } = {}) {

    const { sql, useAnonymousBlock = false, columns = {} } = Parameter.bindSqlParams(sqlString, params);
    let preparedSql = sql;


    let conn = connection;
    const execParams = {
      outFormat: oracledb.OUT_FORMAT_OBJECT
    };
    if(conn) {
      execParams.autoCommit = false;
    } else {
      conn = await Connection.getActiveConnection();
    }

    if (useAnonymousBlock) {
      return this.executeAnonymousBlock(preparedSql, columns);
    }
    try {

      let bindParams = {};

      if (outParams.length) {
        const { sql: sqlWithOutputParams, bind } = Parameter.prepareOutputParams(preparedSql, outParams);

        preparedSql = sqlWithOutputParams;
        bindParams = bind;
      }

      const result = await conn.execute(preparedSql, bindParams, execParams);

      result.connection = conn;

      return result;
    } catch (error) {
      DataBaseLogger.write(error, {sql: sqlString, params, type: 'SQL_EXECUTION'});
      throw error;
    }
  }
}

module.exports = new Oracle();