const oracledb = require('oracledb');

class Parameter {
  static getParamType(type) {
    const types = {
      'number': oracledb.NUMBER,
      'string': oracledb.STRING,
      'cursor': oracledb.CURSOR
    }

    return types[type];
  }

  static create(param) {
    const newParam = {
      dir: Parameter.handleDirectionType(param.dir),
    };
    if(param.maxSize) {
      newParam.maxSize = param.maxSize;
    }
    if(param.type) {
      newParam.type = Parameter.handleParamType(param.type)
    }
    if('value' in param) {
      newParam.val = param.value;
    }
    return newParam;
  }

  static handleDirectionType(direction) {
    const DIRECTIONS = {
      IN: oracledb.BIND_IN,
      OUT: oracledb.BIND_OUT,
      INOUT: oracledb.BIND_INOUT
    };
    return DIRECTIONS[direction] || DIRECTIONS.IN;
  }

  static handleParamType(type) {
    const TYPES = {
      STRING: oracledb.STRING,
      NUMBER: oracledb.NUMBER,
    };
    return type in TYPES ? TYPES[type] : oracledb[type] || null;
  }

  static prepareOutputParams(sql, outParams) {
    const sqlParams = outParams.map((p) => p.name).join(', ');
    const sqlParamsInto = outParams.map((p) => `:${p.name}`).join(', ');
    const sqlReturn = `RETURNING ${sqlParams} INTO ${sqlParamsInto}`;
    const preparedSql = sql + sqlReturn;

    const bind = {};

    for (const param of outParams) {
      bind[param.name] = { type: Parameter.getParamType(param.type), dir: oracledb.BIND_OUT }
    }

    return { sql: preparedSql, bind };
  }

  static bindQueryParams(sql, params) {
    let replacedSql = sql;
    if (params) {
      Object.keys(params).forEach((key) => {
        const replace = new RegExp(`:${key}`, 'g');
        replacedSql = replacedSql.replace(replace, `'${params[key]}'`);
      });
    }
    return replacedSql;
  }

  static bindSqlParams(sql, params) {
    const result = {
      useAnonymousBlock: false,
      sql,
      columns: {},
    };
    const paramsKeys = params ? Object.keys(params) : [];
    paramsKeys.sort((a, b) => (a.length > b.length ? -1 : 1));
    paramsKeys.forEach((key) => {
      try {
        const replace = new RegExp(`:${key}(\\s|,|\\)|$)`, 'g');
        const comma = result.sql.match(replace)[0].includes(',') ? ',' : '';
        const closeParenthesis = result.sql.match(replace)[0].includes(')') ? ')' : '';

        if (params[key] == null) {
          result.sql = result.sql.replace(replace, `null${comma}${closeParenthesis}`);
        } else if (typeof params[key] === 'object') {
          result.useAnonymousBlock = true;
          const colName = key;

          const regExp = new RegExp('"', 'g');
          let paramData = String.fromCharCode(...params[key].data);
          paramData = paramData.replace(regExp, "'");
          columns[colName] = paramData;
        } else if (typeof params[key] === 'string') {
          const regExp = new RegExp("'", 'g');
          params[key] = params[key].replace(regExp, "''");
          result.sql = result.sql.replace(replace, `'${params[key]}'${comma}${closeParenthesis}`);
        } else {
          result.sql = result.sql.replace(replace, `${params[key]}${comma}${closeParenthesis}`);
        }
      } catch (error) {
        // erros que podem acontecer é a chave aparecer no update não como condição e sim para substituir no update
      }
    });
    result.sql = result.sql.replace(/TIMESTAMP null/gi, 'null');
    result.sql = result.sql.replace(/TIMESTAMP 'Invalid date'/gi, 'null');
    return result;
  }
}

module.exports = Parameter;