const apiController = require("../controller/ApiController");
const updaterProdutosModel = require("../model/updaterProdutosModel");
const updaterProdVersaoModel = require("../model/updaterProdVersaoModel");
const updaterModelCasher = require("../model/updaterModelCasher");
const UpdaterServiceMonitoring = require("../provider/UpdaterServiceMonitoring");
const filial = require("../model/Filial");
const caixa = require("../model/Caixa");
const versionCheck = require("../functions/VersionCheck");
const Updater = require("../socket/events/Updater");
const AssistantControl = require("../service/AssistantControl");
const UpdaterEvents = require("../provider/UpdaterEvents");
const UpdaterWarning = require("../helpers/UpdaterWarning");
const DB = require("../database/DB");
const UpdaterHistory = require("../provider/UpdaterHistory");
const UpdaterUnitProduct = require("../provider/UpdaterUnitProduct");
const CasherController = require("../controller/CasherController");
const SystemLogger = require("../logs/SystemLogger");
const UpdaterHandler = require("../socket/UpdaterHandler");

const route = (router) => {
  router.get("/updaterProdutos", async (req, res) => {
    updaterProdutosModel
      .getUpdaterProdutos(true)
      .then((products) => {
        let { rows } = products;
        const { search, search_in } = req.query || {};
        if (search && typeof search_in === "string") {
          rows = rows.filter((row) => {
            return (row[search_in] || "")
              .toString()
              .toLowerCase()
              .includes(search);
          });
        }
        const response = {
          data: rows,
          pagination: {
            limit: 999,
            total: rows.length,
          },
        };
        res.send(response);
      })
      .catch((error) => {
        res.send(error.message || error);
      });
  });

  router.get("/updaterProdutosInfo", async (req, res) => {
    const productInfo = await AssistantControl.getProductInfo();

    if (productInfo) {
      const url = AssistantControl.getControlUrl();

      productInfo.forEach((item) => {
        item.productImage = item.productImage
          ? `${url}/image/${item.productImage}`
          : null;
      });
    }

    res.send(productInfo);
  });

  router.get("/installedVersions", async (req, res) => {
    try {
      const params = {
        UPDATERPRODID: req.query.UPDATERPRODID,
      };
      let version = await apiController.getInstalledVersions(params);

      const nrversao = req.query.NRVERSAO || "";
      if (nrversao) {
        version.rows = version.rows.filter((item) => {
          const checkVersion = versionCheck.greaterThan(
            item.NRVERSAO,
            nrversao
          );
          return checkVersion;
        });
      }

      res.send(version.rows);
    } catch (error) {
      res.status(500).send({
        error: true,
        message: error.message,
      });
    }
  });

  router.get("/installedVersionsCasher", async (req, res) => {
    const params = {
      UPDATERPRODID: req.query.UPDATERPRODID,
    };

    let version = await apiController.getInstalledVersionsCasher(params);

    const nrversao = req.query.NRVERSAO || "";
    if (nrversao) {
      version.rows = version.rows.filter((item) => {
        const checkVersion =
          item.NRVERSAO && versionCheck.greaterThan(item.NRVERSAO, nrversao);
        return checkVersion;
      });
    }

    res.send(version.rows);
  });

  router.get("/state", async (req, res) => {
    let { rows } = await apiController.getState();
    const { search, search_in } = req.query || {};

    if (search && typeof search_in === "string") {
      rows = rows.filter((row) => {
        return (row[search_in] || "").toString().toLowerCase().includes(search);
      });
    }

    const response = {
      data: rows,
      pagination: {
        limit: 999,
        total: rows.length,
      },
    };

    res.send(response);
  });

  router.get("/system-management", async (req, res) => {
    try {
      const {
        PRODUCT_ID,
        CDFILIAL = "",
        SGESTADO = "",
        STATUS = "",
        INITIAL_VERSION = "",
        FINAL_VERSION = "",
      } = req.query;
      if (!PRODUCT_ID) {
        return res.send([]);
      }
      const params = {
        ID: PRODUCT_ID,
        CDFILIAL,
        CDCAIXA: CDFILIAL ? true : false, // provider validation
        SGESTADO,
        STATUS: STATUS === "DESATUALIZADO" ? "" : STATUS,
      };
      
      if (params.CDCAIXA) {
        await CasherController.handleCasherWithoutProduct({ CDFILIAL });
      }

      const latestVersion = await apiController.getLatestProductVersion(
        PRODUCT_ID
      );
      await apiController.setLatestUpdaterApprovedVersion();
      const newestUpdateVersion =
        apiController.getLatestUpdaterApprovedVersion();
      const productInfo = await AssistantControl.getProductInfo(PRODUCT_ID);
      const updaterSockets = await Updater.getUpdaters();
      const products = await UpdaterUnitProduct.getUpdaterProductsSysManagement(
        params
      );
      const connectedUnits = [];
      const connectedCasher = {};

      const unitVersions = {};
      const casherVersions = {};

      updaterSockets.forEach((socket) => {
        const socketVersion = socket.handshake.headers.version;
        const hasUpdate =
          newestUpdateVersion !== "error" &&
          versionCheck.greaterThan(newestUpdateVersion, socketVersion);
        const selfUpdateInfo = {
          UPDATE_AVAILABLE: hasUpdate,
          CURRENT_VERSION: socketVersion,
        };
        if (socket.CDCAIXA) {
          connectedCasher[socket.CDFILIAL] =
          connectedCasher[socket.CDFILIAL] || new Set();
          connectedCasher[socket.CDFILIAL].add(socket.CDCAIXA);
          casherVersions[socket.CDFILIAL] =
          casherVersions[socket.CDFILIAL] || {};
          casherVersions[socket.CDFILIAL][socket.CDCAIXA] = selfUpdateInfo;
        } else {
          connectedUnits.push(socket.CDFILIAL);
          if (newestUpdateVersion !== "error") {
            unitVersions[socket.CDFILIAL] = selfUpdateInfo;
          }
        }
      });

      const checkConnectionStatus = ({ CDFILIAL, CDCAIXA = null }) => {
        if (CDCAIXA) {
          return (
            connectedCasher &&
            connectedCasher[CDFILIAL] &&
            connectedCasher[CDFILIAL].has(CDCAIXA)
          );
        }
        return connectedUnits.includes(CDFILIAL);
      };

      products.rows = await Promise.all(
        products.rows.map(async (product) => {
          const { CDFILIAL: unit, CDCAIXA: casher = null } = product;
          product.REQUISITES = await apiController.checkRequisites(
            product,
            productInfo
          );
          product.CONNECTION_STATUS = checkConnectionStatus({
            CDFILIAL: unit,
            CDCAIXA: casher,
          });
          product.LATEST_VERSION = latestVersion;
          product.WARNING_INFO = product.WARNING
            ? UpdaterWarning.getActiveWarnings({
                unit: product.CDFILIAL,
                casher: product.CDCAIXA || null,
              })
            : [];
          product.WARNING = !!product.WARNING_INFO.length;
          const versionIncompatible =
            product.WARNING && !product.CONNECTION_STATUS;
          if (newestUpdateVersion !== "error") {
            const { CURRENT_VERSION = null, UPDATE_AVAILABLE = null } = casher
              ? (casherVersions[unit] || {})[casher] || {}
              : unitVersions[unit] || {};

            product.UPDATE_AVAILABLE = UPDATE_AVAILABLE;
            product.UPDATE_CURRENT_VERSION = CURRENT_VERSION;
            product.UPDATE_AVAILABLE_VERSION = newestUpdateVersion;
            product.UPDATE_AVAILABLE =
              product.UPDATE_AVAILABLE || versionIncompatible;
          }

          return product;
        })
      );

      if (STATUS === "DESATUALIZADO") {
        products.rows = products.rows.filter((product) => {
          if (product.NRVERSAOATUAL !== "-") {
            return versionCheck.lesserThan(
              product.NRVERSAOATUAL,
              product.LATEST_VERSION
            );
          }
        });
      }

      products.rows = products.rows.filter((product) => {
        if (
          product.NRVERSAOATUAL === "-" &&
          (INITIAL_VERSION || FINAL_VERSION)
        ) {
          return false;
        }
        const currentVersion = versionCheck.parse(product.NRVERSAOATUAL);
        const isValidByMinVersion = INITIAL_VERSION
          ? versionCheck.greaterThanEqual(currentVersion, INITIAL_VERSION)
          : true;
        const isValidByMaxVersion = FINAL_VERSION
          ? versionCheck.lesserThanEqual(currentVersion, FINAL_VERSION)
          : true;
        return isValidByMinVersion && isValidByMaxVersion;
      });

      if (products.rows.length === 0) {
        return res.send({
          error: true,
          message: "Nenhum registro encontrado.",
        });
      }
      const response = {
        data: products.rows,
        pagination: {
          limit: 999,
          total: products.rows.length,
        },
      };

      res.send(response);
    } catch (error) {
      SystemLogger.error(error?.message + `\n${error.stack}`);
      res.send(error.message || error);
    }
  });

  router.post("/updaterProductMultiInstall", async (req, res) => {
    const { PRODUCT_ID, UNITS, UNITS_AND_CASHER = [] } = req.body;
    const useCasher = !!UNITS_AND_CASHER.length;
    if (!PRODUCT_ID) {
      res.send([]);
      return;
    }
    try {
      let productVersions = { rows: [] };
      if (useCasher) {
        productVersions =
          await updaterProdutosModel.getUpdaterProductsWithVersionAndCasher(
            PRODUCT_ID
          );
      } else {
        productVersions =
          await updaterProdutosModel.getUpdaterProductsWithVersion(PRODUCT_ID);
      }
      const updaterSockets = await Updater.getUpdaters();
      const connectedUnits = [];
      const connectedUnitsAndCasher = [];
      updaterSockets.forEach((socket) => {
        if (socket.CDCAIXA) {
          connectedUnitsAndCasher.push(`${socket.CDFILIAL}_${socket.CDCAIXA}`);
        } else {
          connectedUnits.push(socket.CDFILIAL);
        }
      });
      const onlySelectedUnits = ({ CDFILIAL }) => UNITS.includes(CDFILIAL);
      const onlySelectedCasher = ({ CDFILIAL, CDCAIXA }) =>
        UNITS_AND_CASHER.includes(`${CDFILIAL}_${CDCAIXA}`);
      const fillConnectionStatus = (row) => {
        let isConnected = false;
        if (useCasher) {
          isConnected = connectedUnitsAndCasher.includes(
            `${row.CDFILIAL}_${row.CDCAIXA}`
          );
        } else {
          isConnected = connectedUnits.includes(row.CDFILIAL);
        }
        row.CONNECTION_STATUS = isConnected;
      };
      productVersions.rows = productVersions.rows.filter(
        useCasher ? onlySelectedCasher : onlySelectedUnits
      );
      productVersions.rows.forEach(fillConnectionStatus);
      const response = {
        data: productVersions.rows,
        pagination: {
          limit: 999,
          total: productVersions.rows.length,
        },
      };
      res.send(response);
    } catch (error) {
      res.status(500).send(error.message || error);
    }
  });

  router.post("/getEventsInProgress", async (req, res) => {
    const { CDFILIAL, UNITS = [], CASHER = [] } = req.body;
    try {
      const params = {};
      const useCasher = Array.isArray(CASHER) && CASHER.length;
      if (useCasher) {
        params.P_CDFILIAL = CDFILIAL;
        params.P_CDCAIXA = `_${CASHER.join("_")}_`;
      } else {
        params.P_CDFILIAL = `_${UNITS.join("_")}_`;
      }
      const EVENTS_IN_PROGRESS = await UpdaterEvents.getEventsInProgress(
        params
      );
      res.send({
        EVENTS: EVENTS_IN_PROGRESS.rows,
      });
    } catch (error) {
      res.status(500).send(error.message || error);
    }
  });

  router.get("/updaterParametrizationCashers", async (req, res) => {
    try {
      const { CDFILIAL } = req.query;
      const params = { CDFILIAL };

      const cashers = await caixa.getActiveCashersFromUnit(params);

      const socketUpdaters = await Updater.getUpdaters();
      const connectedFiliais = socketUpdaters.reduce((acc, socket) => {
        acc[socket.CDFILIAL] = acc[socket.CDFILIAL] || new Set();
        if (socket.CDCAIXA) {
          acc[socket.CDFILIAL].add(socket.CDCAIXA);
        }
        return acc;
      }, {});

      const unitsWithWarning = UpdaterWarning.getMappedUnitsCashersWithWarning();

      cashers.rows = await Promise.all(
        cashers.rows.map(async (casher) => {
          const curUnit = connectedFiliais[casher.CDFILIAL] || new Set();
          const cashersWithWarning = unitsWithWarning[casher.CDFILIAL] || [];
          casher.CONNECTION_STATUS = curUnit.has(casher.CDCAIXA);
          casher.WARNING = cashersWithWarning.includes(casher.CDCAIXA);
          return casher;
        })
      );

      const response = {
        data: cashers.rows,
        pagination: {
          limit: 999,
          total: cashers.rows.length,
        },
      };

      res.send(response);
    } catch (error) {
      res.send(error.message || error);
    }
  });

  router.get("/updaterProdutosFilialPerCasher", async (req, res) => {
    const productId = req.query.PRODUCT_ID;
    const latestVersion = await apiController
      .getLatestProductVersion(productId)
      .catch((err) => null);
    const productInfo = await AssistantControl.getProductInfo(productId);
    updaterProdutosModel
      .getUpdaterProdutosWithUnitsPerCasher(req.query)
      .then(async (products) => {
        const casherAndProducts =
          await updaterModelCasher.getCasherWithoutProduct({
            CDFILIAL: req.query.CDFILIAL,
            UPDATERPRODID: req.query.PRODUCT_ID,
          });
        if (
          casherAndProducts &&
          casherAndProducts.rows &&
          casherAndProducts.rows.length
        ) {
          await Promise.all(
            casherAndProducts.rows.map((rowToInsert) => {
              return UpdaterUnitProduct.insert(rowToInsert);
            })
          ).catch((err) => {
            DB.rollback();
          });
          products =
            await updaterProdutosModel.getUpdaterProdutosWithUnitsPerCasher(
              req.query
            );
        }
        let socketUpdaters = await Updater.getUpdaters();
        const connectedFiliais = socketUpdaters.reduce((acc, socket) => {
          acc[socket.CDFILIAL] = acc[socket.CDFILIAL] || new Set();
          if (socket.CDCAIXA) {
            acc[socket.CDFILIAL].add(socket.CDCAIXA);
          }
          return acc;
        }, {});
        const unitsWithWarning =
          UpdaterWarning.getMappedUnitsCashersWithWarning();
        products.rows = await Promise.all(
          products.rows.map(async (product) => {
            const curUnit = connectedFiliais[product.CDFILIAL] || new Set();
            const cashersWithWarning = unitsWithWarning[product.CDFILIAL] || [];
            product.REQUISITES = await apiController.checkRequisites(
              product,
              productInfo
            );
            product.CONNECTION_STATUS = curUnit.has(product.CDCAIXA);
            product.WARNING = cashersWithWarning.includes(product.CDCAIXA);
            product.LATEST_VERSION = latestVersion;
            return product;
          })
        );
        const response = {
          data: products.rows,
          pagination: {
            limit: 999,
            total: products.rows.length,
          },
        };
        res.send(response);
      })
      .catch((error) => {
        res.send(error.message || error);
      });
  });

  router.get("/updaterProdutosVersao", async (req, res) => {
    const productId = req.query.CDPRODUCTID;
    const { search } = req.query || {};
    if (!productId) {
      res.send([]);
    } else {
      updaterProdVersaoModel
        .getUpdaterProdVersaoByProduct(productId)
        .then((products) => {
          let { rows } = products;
          if (search) {
            rows = rows.filter((row) => {
              return (row.NRVERSAO || "").includes(search);
            });
          }

          rows.sort((x, y) => versionCheck.sort(x.NRVERSAO, y.NRVERSAO));
          const response = {
            data: rows,
            pagination: {
              limit: 999,
              total: products.rows.length,
            },
          };
          res.send(response);
        })
        .catch((error) => {
          res.send(error.message || error);
        });
    }
  });

  router.get("/updaterProdutosVersaoConfig", async (req, res) => {
    try {
      const products =
        await updaterProdVersaoModel.getUpdaterProdVersaoConfig();

      const sortedProducts = products.rows.sort((x, y) => {
        return versionCheck.sort(x.NRVERSAO, y.NRVERSAO);
      });
      const response = {
        data: sortedProducts,
        pagination: {
          limit: 999,
          total: sortedProducts.length,
        },
      };
      res.send(response);
    } catch (error) {
      res.send(error.message || error);
    }
  });

  router.get("/updaterProdutoHistorico", async (req, res) => {
    const { CDFILIAL, CDPRODUCTID } = req.query;
    UpdaterHistory.getProductHistory({ CDFILIAL, CDPRODUCTID })
      .then((historico) => {
        const response = {
          data: historico.rows,
          pagination: {
            limit: 999,
            total: historico.rows.length,
          },
        };
        res.send(response);
      })
      .catch((error) => {
        res.send(error.message || error);
      });
  });

  router.get("/updaterProdutoHistoricoPorCaixa", async (req, res) => {
    const { CDFILIAL, CDCAIXA, CDPRODUCTID } = req.query;
    UpdaterHistory.getProductHistory({ CDFILIAL, CDCAIXA, CDPRODUCTID })
      .then((historico) => {
        const response = {
          data: historico.rows,
          pagination: {
            limit: 999,
            total: historico.rows.length,
          },
        };
        res.send(response);
      })
      .catch((error) => {
        res.send(error.message || error);
      });
  });

  router.put("/updateUpdaterProdVersaoAprov", async (req, res) => {
    try {
      await updaterProdVersaoModel.updateUpdaterProdVersaoAprov({
        ID: req.body.ID,
        IDAPROVVER: req.body.IDAPROVVER,
      });

      res.send("Sucesso!");
    } catch (error) {
      res.send(error.message || error);
    }
  });

  router.get("/updaterUnits", async (req, res) => {
    try {
      let { rows } = await filial.getActiveUnits();
      const { search, search_in } = req.query || {};
      if (search && typeof search_in === 'string') {
        rows = rows.filter(row => {
            return ( row[search_in] || "" )
            .toString()
            .toLowerCase()
            .includes(search);
        });
      }

      const response = {
        data: rows,
        pagination: {
          limit: 999,
          total: rows.length,
        },
      };

      res.send(response);
    } catch (error) {
      res.send(error.message || error);
    }
  });

  router.get('/monitored-units', async(req, res) => {
    try {
      const { search, casher = null } = req.query || {};
      const { rows } = await UpdaterServiceMonitoring.getMonitoredUnits({ CDCAIXA: casher });
      let filteredRows = rows;
      if (search) {
        const valueToSearch = search.toLowerCase();
        filteredRows = rows.filter(({ CDFILIAL, NMFILIAL }) => {
          const unit = `${CDFILIAL} - ${NMFILIAL}`.toLowerCase();
          return unit.includes(valueToSearch);
        });
      }

      const response = {
        data: filteredRows,
        pagination: {
          limit: 999,
          total: filteredRows.length,
        },
      };

      res.send(response);
    } catch (error) {
      res.status(500).send(error.message || error);
    }
  });

  router.get("/service-management", async (req, res) => {
    try {
      const params = {
        P_CDFILIAL: req.query.CDFILIAL,
        P_CDCAIXA: req.query.CDCAIXA || null,
      }
      const { rows } = await UpdaterServiceMonitoring.getServiceMonitoring(params);

      if (!rows.length) throw "Nenhum serviço encontrado.";
      const data = [];
      const socketInfo = await UpdaterHandler.getMapConnectedSockets();

      rows.forEach(row => {
        const { CDFILIAL: unit, CDCAIXA: casher = null } = row;
        const newRow = {
          CDFILIAL: unit,
          CONNECTION_STATUS: false,
        }
        if(casher) {
          newRow.CDCAIXA = casher;
        }
        const onlyUnit = !casher && socketInfo[unit] && !socketInfo[unit].withCasher;
        const hasUnitAndCasher = casher && socketInfo[unit] && socketInfo[unit][casher];
        if (onlyUnit || hasUnitAndCasher) {
          const warningInfo = casher ? socketInfo[unit][casher] : socketInfo[unit];
          Object.assign(newRow, warningInfo, { CONNECTION_STATUS: true });
        }
        row.MONITORAMENTO.split(';').forEach(serviceName => {
          if (serviceName) {
            data.push({
              ...newRow,
              ID: `${serviceName}_${row.CDFILIAL}` + (row.CDCAIXA ? `_${row.CDCAIXA}`: ''),
              NMSERVICO: serviceName,
            });
          }
        });
      });

      const response = {
        data,
        pagination: {
          limit: 999,
          total: rows.length,
        },
      };
      res.send(response);
    } catch (error) {
      res.status(500).send(error.message || error);
    }
  })

  router.get("/minimumSupportedVersion", async (req, res) => {
    const { MIN_UPDATER_VERSION_SUPPORTED } = process.env;
    res.send({ minimumSupportedVersionUpdater: MIN_UPDATER_VERSION_SUPPORTED });
  });

  router.get("/service-management-casher", async (req, res) => {
    try {
      const { search } = req.query || {};
      const defaultSelectedCasher = { CDCAIXA: "T", CAIXA: "Todos" };
      let { rows } = await UpdaterServiceMonitoring.getServiceMonitoringWithCasher(req.query);

      if (search) {
        rows = rows.filter((row) => {
          return (row.FILIAL || "").includes(search);
        });
      }

      if (!rows.length) throw "Nenhum caixa encontrado.";

      rows.unshift(defaultSelectedCasher);
      const response = {
        data: rows,
        pagination: {
          limit: 999,
          total: rows.length,
        },
      };
      res.send(response);
    } catch (error) {
      res.status(500).send(error.message || error);
    }
  });
};

module.exports = route;
