const ms = require('ms');
const AssistantConfig = require("../../model/AssistantConfig");
const UpdaterServicoFili = require("../../model/UpdaterServicoFili");
const SocketRooms = require("../SocketRooms");
const ServicesMonitoring = require("../../helpers/ServicesMonitoring");
const Gzip = require('../../functions/Gzip');
const SystemLogger = require("../../logs/SystemLogger");
const Filial = require("../../model/Filial");
const DB = require('../../database/DB');
const SocketManager = require('../SocketManager');
const MonitoringData = require('../MonitoringData');
const MonitoringInterval = require('../MonitoringInterval');

class Monitor {
    static monitoringPageName = SocketRooms.master.monitoringPage;
    static maxInactivityTime = ms('10m')
    static timeLastConnectedUser = null;
    static inactiveUserInterval = setInterval(() => {
        const hasActiveMonitoring = MonitoringInterval.hasSomeActive();
        const hasActiveUser = Monitor.timeLastConnectedUser === null;
        const timeLapseLastUser = Date.now() - (Monitor.timeLastConnectedUser || Date.now());
        if(hasActiveMonitoring && !hasActiveUser && timeLapseLastUser >= Monitor.maxInactivityTime) {
            MonitoringInterval.clear();
        }
    }, Monitor.maxInactivityTime);
    static unitsWithSat = new Set();
    monitoringParams = {};
    intervalLoadUnitsWithSat = null;

    async register(socket, { cdfilial, cdcaixa }) {
        socket.CDFILIAL = cdfilial;
        socket.CDCAIXA = cdcaixa;
        const hasConsumers = await this.hasConsumersForMonitoring();
        socket.join(SocketRooms.client.default);

        if(hasConsumers) {
            this.startMonitoring([socket]);
            this.getMonitorRoom().emit('monitorConnected', { status: 'S', unit: cdfilial });
        }
        socket.on(`MONITORING_RESULT_APACHE`, (result) => {
            this.handleResponseMonitor(result);
        });

        socket.on(`MONITORING_RESULT_LOCAL_DB`, (result) => {
            this.handleResponseMonitor(result, { findErrorMethod: () => false });
        });

        socket.on(`MONITORING_RESULT_LOCAL_SERVICES`, (result) => {
            this.handleResponseMonitor(result);
        });

        socket.on("MONITORING_RESULT_SAT", (result) => {
            this.handleResponseMonitor(result);
        });
    }

    handleResponseMonitor(result, { findErrorMethod = null } = {}) {
        try {
            const verifyHasError = (data) => {
                if(!data || !data.response || !Array.isArray(data.response.rows)) return true;
                return data.response.rows.some(row => row.STATUS === 'N');
            }
            const executionError = result.isError;
            const { ID, NOME, CDFILIAL, TIPO, COMPLEMENTOTIPO } = result.data;
            const config = { ID, NOME, CDFILIAL, TIPO, COMPLEMENTOTIPO };
            const testFunction = typeof findErrorMethod === 'function' ? findErrorMethod : verifyHasError;
            if (executionError) {
                const errorResponse = {
                    metadata: [{ name: "ERRO" }],
                    rows: [{ ERRO: result.response }],
                };
                this.setMonitoringResult(config, errorResponse, true);
                return;
            }
            const resultClone = JSON.parse(JSON.stringify(result));
            const hasError = testFunction(result);
            if (!hasError) {
                resultClone.response.rows = [];
            }
            this.setMonitoringResult(config, resultClone.response);
        } catch(error) {
            SystemLogger.write(error, { type: 'OCCURRED_ERROR_HANDLE_MONITOR_DATA' });
        }
    }
    unregister(socket, { cdfilial }) {
        this.getMonitorRoom().emit('monitorConnected', { status: 'N', unit: cdfilial });
    }
    setTimeLastConnectedUser(time) {
        Monitor.timeLastConnectedUser = time;
    }
    getMonitorRoom() {
        return SocketManager.getRoom(Monitor.monitoringPageName);
    }
    async execCommand(command, params) {
        command = `/* === EXECUTED BY ODHEN ASSISTANT === */ \n ${command}`;
        const { rows, metaData } = await DB.execute(command, params);
        return { rows, metaData };
    }
    async monitoringDataBaseCloud(config) {
        const { CDFILIAL, COMANDO } = config;
        let hasError = false;
        let response = null;
        try {
            response = await this.execCommand(COMANDO, { CDFILIAL })
        } catch (error) {
            response = {
                metaData: [{ name: "ERRO" }],
                rows: [[error.message]],
            };
            hasError = true;
        }
        this.setMonitoringResult(config, response, hasError);
    }
    async execHttpRequest(ID, urls) {
        return Promise.all(urls.map((url) => {
          return new Promise(async(resolve) => {
            const response = {
              ID,
              URL: url,
              ERROR: '',
              STATUS: ''
            };
            try {
              await axios.get(url.trim(), {}, { timeout: ms('10s') });
              response.STATUS = 'S';
            } catch (error) {
              response.STATUS = 'N';
              response.ERROR = error.message;
            } finally {
              resolve(response);
            }
          });
        }));
    }

    async hasConsumersForMonitoring(){
        const consumers = await this.getMonitorRoom().fetchSockets();
        return consumers.length > 0;
    }

    async loadInitialParams(){
        return Promise.all([
            this.loadUnitsWithSat(),
            this.loadMonitoringParams(),
        ])
    }

    connectedUser(){
        Monitor.timeLastConnectedUser = null;
        this.startMonitoring();
    }
    async disconnectedUser() {
        const usersWithMonitoringActive = await this.getMonitorRoom().fetchSockets();
        if(usersWithMonitoringActive.length === 0) {
            Monitor.timeLastConnectedUser = Date.now();
        }
    }
    getConfigById(ID, paramsToInclude = {}) {
        return ({...this.monitoringParams[ID], ...paramsToInclude});
    }
    getNewUnitsWithSat(updatedUnits){
        return [...updatedUnits].filter(unit => !Monitor.unitsWithSat.has(unit));
    }
    async setUnitsWithSat( unitsWithSat, updateMonitoring = false ) {
        if(updateMonitoring) {
            const newUnitsWithSat = this.getNewUnitsWithSat(unitsWithSat);
            if(newUnitsWithSat.length) {
                const clientSockets = await this.getSocketsClientByUnits(newUnitsWithSat);
                const satMonitoring = this.getMonitorsParams().find(config => config.COMANDO === 'INTERNO' && config.IDMONITORAMENTO === 'SAT');
                Monitor.unitsWithSat = unitsWithSat;
                this.startMonitoring(clientSockets, satMonitoring.ID);
            }
        } else {
            Monitor.unitsWithSat = unitsWithSat;
        }

    }
    async fetchUnitsWithSat(){
        const queryResultAll = await Filial.getUnits();
        const { rows } = queryResultAll;
        const unitsWithSat = new Set();
        rows.forEach(({SAT, CDFILIAL}) => {
            if( SAT !== 0 ) unitsWithSat.add(CDFILIAL);
        });
        return unitsWithSat;
    }
    async loadUnitsWithSat() {
        this.setUnitsWithSat(await this.fetchUnitsWithSat());
        this.intervalLoadUnitsWithSat = setInterval(() => {
            this.fetchUnitsWithSat().then( units => this.setUnitsWithSat(units, true) );
        }, ms('1d'));
    }
    isUnitWithSat( CDFILIAL ) {
        return Monitor.unitsWithSat.has(CDFILIAL);
    }

    getMonitorsParams(ID = null) {
        if( !ID ) return Object.values(this.monitoringParams);
        if( this.monitoringParams[ID] ) return [this.monitoringParams[ID]];
        return [];
    }

    getUnitBySocketClient( socket ) {
        const { headers = {} } = socket.handshake;
        const { cdfilial = 'Unidade não informada no socket-client' } = headers;
        return cdfilial;
    }

    isMonitoringServiceInUnit( ID = null ) {
        const {
            TIPO,
            IDMONITORAMENTO,
        } = this.monitoringParams[ID] || {};
        return (ID && TIPO === 'I' && IDMONITORAMENTO === 'SERVICOS_UPDATER');
    }

    async startMonitoring( clients = null, ID = null ){
        const hasActiveMonitoring = MonitoringInterval.hasSomeActive();
        const isNewClientConnection = !!clients;
        const isUpdateMonitoringParams = !!ID;
        const needCreateOrUpdateMonitoring = isNewClientConnection
                                          || isUpdateMonitoringParams
                                          || !hasActiveMonitoring;
        if(!needCreateOrUpdateMonitoring) {
            const allClients = await this.getMonitors();
            const clientsWithoutMonitoring = allClients.filter(socket => {
                return !MonitoringInterval.hasActive(null, { unit: socket.CDFILIAL, casher: socket.CDCAIXA });
            });
            if(!clientsWithoutMonitoring.length) {
                return;
            }
            clients = clientsWithoutMonitoring;
        }
        clients = Array.isArray(clients) ? clients : await this.getMonitors();
        clients.forEach((socketClient) => {
            let monitorsToExecute = this.getMonitorsParams(ID);
            monitorsToExecute.forEach(config => {
                const ID = config.ID;
                const CDFILIAL = this.getUnitBySocketClient(socketClient);
                //? A => Execução na base cloud do Assistant
                //? I => Execução de rotinas criadas internamente que não podem ser alteradas via sistema (SAT, por exemplo)
                //? L => Execução na base Lojas, envia a requisição e aguarda o retorno para emitir para o Assistant
                //? E => Execução de requisição HTTP nas Lojas, envia a requisição e aguarda o retorno para emitir para o Assistant
                //? S => Verificação se Serviços do Windows na Loja estão em execução
                const functionPerType = {
                    A: this.monitoringDataBaseCloud,
                    I: this.monitoringAssistantInterno,
                    L: (config, socket) => socket.emit('execMonitoringSQL', config),
                    E: (config, socket) => socket.emit('execMonitoringHttp', config),
                    S: (config, socket) => socket.emit('execMonitoringServices', config),
                };
                const functionToCall = functionPerType[config.TIPO] || (() => {});
                const conf = this.getConfigById(ID, { CDFILIAL });
                functionToCall.apply(this, [conf, socketClient]);
                const intervalTimeMs = config.INTERVALO * 1000;
                MonitoringInterval.remove(config.ID, { unit: CDFILIAL });
                const callbackFn = () => {
                    functionToCall.apply(this, [this.getConfigById(ID, { CDFILIAL }), socketClient]);
                }
                MonitoringInterval.add(config.ID, callbackFn, intervalTimeMs, { unit: CDFILIAL });
            })
        });
    }
    hasResult( dataset ) {
        return dataset && dataset.rows && dataset.rows.length;
    }
    async loadMonitoringParamsServicesInUnit( ID ){
        const updaterServicoFili = await UpdaterServicoFili.getAllServiceMonitoring();
        if(ID && this.hasResult(updaterServicoFili)) {
            const servicesInUnits = updaterServicoFili.rows;
            this.monitoringParams[ID].UNITS = servicesInUnits.reduce((acc, cv) => {
                acc[cv.CDFILIAL] = {COMANDO: cv.MONITORAMENTO};
                return acc;
            }, {});
        }
    }
    async loadMonitoringParams(){
        try {
            const monitoringParams = await AssistantConfig.getMonitoringParams();

            let idServicoFiliMonitor = null;
            if(this.hasResult(monitoringParams)) {
                this.monitoringParams = monitoringParams.rows.reduce((acc, cv) => {
                    acc[cv.ID] = {...cv};
                    if(cv.TIPO === 'I' && cv.IDMONITORAMENTO === 'SERVICOS_UPDATER') idServicoFiliMonitor = cv.ID;
                    if(cv.TIPO !== 'I' && cv.COMANDO) {
                        acc[cv.ID].COMANDO = Gzip.decode(cv.COMANDO);
                    }
                    return acc;
                }, {});
            }
            await this.loadMonitoringParamsServicesInUnit(idServicoFiliMonitor);
            this.startMonitoring();
        } catch (error) {
            return Promise.reject(error);
        }
    }

    async getSocketsClient(){
        return await SocketManager.getSocketIo().in("ASSISTANT_CLIENT_ROOM").fetchSockets();
    }

    async getSocketClientByUnit(CDFILIAL) {
        return this.getSocketsClient()
            .then((allSockets) => {
                const clientUnitSocket = allSockets.find((socket) => {
                    return socket.CDFILIAL == CDFILIAL;
                });
                return clientUnitSocket;
            });
    }
    async getSocketsClientByUnits( UNITS ) {
        const sockets = await this.getSocketsClient();
        return sockets.filter( socket => UNITS.includes(socket.CDFILIAL));
    }
    setSocketIo(io) {
        this.io = io;
    }

    async getMonitors() {
        return await SocketManager.getRoom(SocketRooms.client.default).fetchSockets();
    }
    updateMonitoringParams(config) {
        const {ATIVO, ...monitorParams} = config;
        if(ATIVO == 'N') {
            return MonitoringInterval.removeById(config.ID);
        }
        if(config.TIPO === 'S') {
            monitorParams.COMANDO = ServicesMonitoring.removeMonitoringParamInconsistencies(monitorParams.COMANDO);
        }
        this.monitoringParams[config.ID] = this.monitoringParams[config.ID] || {};
        Object.assign(this.monitoringParams[config.ID], monitorParams);
        this.hasConsumersForMonitoring().then(hasConsummers => {
            if(hasConsummers || MonitoringInterval.hasSomeActive()) {
                this.startMonitoring(null, config.ID);
            }
        });
    }

    setMonitoringResult(config, response, isExecutionError = false) {
        const {
            ID,
            TIPO,
            COMPLEMENTOTIPO,
            CDFILIAL,
            CDCAIXA = null,
        } = config;

        const active = response.rows.every( item => item.STATUS === 'S');
        let QTDERROR = 0;
        let STATUS = '';

        for(const item of response.rows) {
            if(!item.STATUS) {
                QTDERROR = response.rows.length;
                break;
            }
            if(item.STATUS === 'N') QTDERROR++;
        };

        switch (TIPO) {
            case 'A':
            case 'L':
                if (COMPLEMENTOTIPO === 'S') {
                    STATUS = response.rows.length > 0 ? 'S' : 'N';
                } else {
                    STATUS = response.rows.length > 0 ? 'N' : 'S';
                }
                break;
            default:
                STATUS = active ? 'S' : 'N';
        }
        if(isExecutionError) STATUS = 'N';
        if(STATUS === 'S') QTDERROR = 0;
        const monitorResult = {
            ...response,
            STATUS,
            QTDERROR,
        }
        MonitoringData.add(monitorResult, {
            id: ID,
            unit: CDFILIAL,
            caasher: CDCAIXA,
        })
        //Envia informações atualizadas para a tela de monitoramento.
        this.getMonitorRoom().emit(ID, {ID, CDFILIAL, STATUS, QTDERROR});
    }

    monitoringSat(config, socket) {
        const { CDFILIAL } = config;
        if (this.isUnitWithSat(CDFILIAL)) {
            return socket.emit('satMonitoring', config);
        }
        MonitoringInterval.remove(config.ID, { unit: CDFILIAL });
    }

    monitoringServicesByUnit(config, socket) {
        const { ID, CDFILIAL, UNITS = {}, ...configMonitor } = config;
        const { COMANDO } = (UNITS[CDFILIAL] || {});
        if(ID && COMANDO) {
            const newConfig = {...configMonitor, COMANDO, CDFILIAL, ID};
            socket.emit("execMonitoringServices", newConfig);
        }
    }
    monitoringAssistantInterno(config, socket) {
        switch( config.IDMONITORAMENTO ) {
            case 'SAT':
                this.monitoringSat(config, socket);
                break;
            case 'SERVICOS_UPDATER':
                this.monitoringServicesByUnit(config, socket);
                break;
            default:
                SystemLogger.write(`INVALID_INTERNAL_MONITORING: ${config.NOME}`);
        }
    }
    getServicesUpdaterMonitor( CDFILIAL ){
        let { ID, UNITS } = Object.values(this.monitoringParams).find( monitor => {
            return monitor.TIPO === 'I' && monitor.IDMONITORAMENTO === 'SERVICOS_UPDATER';
        }) || {ID: null, UNITS: {}};
        if(ID && !UNITS) {
            this.monitoringParams[ID].UNITS = {};
            UNITS = this.monitoringParams[ID].UNITS;
        }
        UNITS[CDFILIAL] = UNITS[CDFILIAL] || {COMANDO: ''};
        const { COMANDO } = UNITS[CDFILIAL];
        return {
            ID,
            COMANDO,
            PARAMS: UNITS[CDFILIAL]
        };
    }
    updateServiceToMonitoring(CDFILIAL, services, operation) {
        const { ID, COMANDO, PARAMS } = this.getServicesUpdaterMonitor( CDFILIAL );
        if( !ID ) return;
        let actualServices = COMANDO.length ? new Set(COMANDO.split(';')) : new Set();
        const addService = service => actualServices.add(service);
        const deleteService = service => actualServices.delete(service);
        services.split(';').forEach(operation === 'add' ? addService : deleteService);
        PARAMS.COMANDO = [...actualServices].join(';');
        return { ID, COMANDO: PARAMS.COMANDO, UPDATED: COMANDO !== PARAMS.COMANDO };
    }
    async addServiceToMonitoring(services, { unit }) {
        try {
            const { ID, UPDATED } = this.updateServiceToMonitoring(unit, services, 'add') || {};
            if(!ID) return;
            if( UPDATED && !MonitoringInterval.hasActive(ID, { unit })) {
                const socketClient = await this.getSocketClientByUnit(unit);
                if(socketClient) {
                    this.startMonitoring([socketClient], ID);
                }
            }
        } catch( error ) {
            SystemLogger.write(error, { type: 'ERR_ADD_MONITORING' });
        }
    }
    removeServiceToMonitoring(services, { unit, casher }) {
        try {
            const { ID, COMANDO, UPDATED } = this.updateServiceToMonitoring(unit, services, 'delete') || {};
            if(!ID) return;
            if(UPDATED && !COMANDO && MonitoringInterval.hasActive(ID, { unit })) {
                MonitoringInterval.remove(ID, { unit, casher });
                MonitoringData.remove({ id: ID, unit, casher });
                const socketData = {
                    ID,
                    CDFILIAL: unit,
                    STATUS: false,
                };
                if (casher) {
                    socketData.CDCAIXA = casher;
                }
                this.getMonitorRoom().emit(ID, socketData);
            }
        } catch( error ) {
            SystemLogger.write(error, { type: 'ERR_REMOVE_MONITORING' });
        }
    }
}

module.exports = new Monitor();